#include "BAMWriter.h"

#ifdef _BAM

#include <string.h>

#include "api/BamWriter.h"
#include "api/SamHeader.h"
#include "api/BamAlignment.h"

#include "Config.h"
#include "SequenceProvider.h"
#include "NGM.h"

using namespace BamTools;

//using BamTools::BamWriter;
//using BamTools::SamHeader;
//Format: http://samtools.sourceforge.net/SAM1.pdf

//static inline void rev(char * s);

BamWriter writer;

void BAMWriter::DoWriteProlog() {
	//TODO: check correct format	;

	SamHeader header;
	RefVector refs;

	SamProgram program;

	header.Version = "1.0";
	header.SortOrder = "unsorted";

	program.Name = "ngm";
	program.Version = "0.0.1";
	program.CommandLine = std::string(Config.GetString("cmdline"));

	header.Programs.Add(program);

	char const * refName = 0;
	int refNameLength = 0;
	int ref = Config.GetInt("ref_mode");

	if (ref == -1)
		for (int i = 0; i < SequenceProvider.GetRefCount(); ++i) {
			refName = SequenceProvider.GetRefName(i, refNameLength);
			//Print("@SQ\tSN:%.*s\tLN:%d\n", refNameLength, refName, SequenceProvider.GetRefLen(i));
			RefData bamRef(std::string(refName, refNameLength), SequenceProvider.GetRefLen(i));
			refs.push_back(bamRef);
			if (NGM.DualStrand())
				++i;
		}
	else {
		refName = SequenceProvider.GetRefName(ref * ((NGM.DualStrand()) ? 2 : 1), refNameLength);
		RefData bamRef(std::string(refName, refNameLength), SequenceProvider.GetRefLen(ref * ((NGM.DualStrand()) ? 2 : 1)));
		refs.push_back(bamRef);
	}

	if (!writer.Open(file, header, refs)) {
		Log.Error("Could not open output BAM file");
		return;
	}
}

void BAMWriter::DoWritePair(MappedRead const * const read1, MappedRead const * const read2) {
	throw "not yet implemented (bam paired output";
}

void BAMWriter::DoWriteRead(MappedRead const * const read) {
	BamAlignment al;

	char const * readseq = read->Seq;
	int readlen = read->length;

	char * readname = read->name;
	int readnamelen = read->nameLength;

	char * qltystr = read->qlty;
	int qltylen = readlen;

	al.AlignmentFlag = 0;
	al.SetIsMapped(true);

	if ((read->Strand == '-')) {
		readseq = read->RevSeq;
		//TODO: also reverse quality values???
		al.SetIsReverseStrand(true);
	}

	al.Name = std::string(readname, readnamelen);
	al.Length = readlen;
	al.MapQuality = read->mappingQlty;
	al.Position = read->TLS()->Location.m_Location;
	std::string seq(readseq, readlen);
	al.QueryBases = seq;
	al.RefID = read->TLS()->Location.m_RefId / 2;

	int length = atoi(read->Buffer1);
	int i = 0;
	Log.Message("%s: %s", read->name, read->Buffer1);
	while (read->Buffer1[i] != '\0') {
		if (read->Buffer1[i] < '0' || read->Buffer1[i] > '9') {
			al.CigarData.push_back(CigarOp(read->Buffer1[i], length));
			length = atoi(read->Buffer1 + i + 1);
		}
		i += 1;
	}

	if (qltystr != 0 && qltylen > 0 && qltylen == readlen) {
		al.Qualities = std::string(qltystr, qltylen);
	} else {
		al.Qualities = std::string(readlen, ':');
	}

	//	//Optional fields
	al.AddTag("AS", "i", (int) read->TLS()->Score.f);
	al.AddTag("MD", "Z", std::string(read->Buffer2));
	al.AddTag("X0", "i", (int) read->EqualScoringCount);
	al.AddTag("XI", "f", read->Identity);

	writer.SaveAlignment(al);
}

void BAMWriter::DoWriteUnmappedRead(MappedRead const * const read) {

}

void BAMWriter::DoWriteEpilog() {
	writer.Close();
}

#endif
