#ifndef __GENERICREADWRITER_H__
#define __GENERICREADWRITER_H__

#include <stdio.h>
#include <stdarg.h>

#include "ILog.h"
#include "Config.h"

#include "MappedRead.h"

class GenericReadWriter {
public:
	GenericReadWriter(char const * const filename) {
		if (!(m_Output = fopen(filename, "wb"))) {
			Log.Error("Unable to open output file %s", filename);
		}
		writeBuffer = new char[BUFFER_SIZE];
		bufferPosition = 0;
		if (Config.Exists("paired")) {

			paired_end = (Config.GetInt("paired", 0, 1) == 1);
			if (Config.Exists("pair_min_distance") && Config.Exists("pair_max_distance")) {
				min_dist = Config.GetInt("pair_min_distance");
				max_dist = Config.GetInt("pair_max_distance");
			} else {
				Log.Warning("No paired end distance is set.");
				min_dist = 0;
				max_dist = 999999999;
			}
		}
		identity = 0;
		if (Config.Exists("identity_tresh")) {
			identity = Config.GetFloat("identity_tresh", 0, 100);
			identity = identity / 100.0f;
		}
	}
	virtual ~GenericReadWriter() {
		if (m_Output) {
			Flush(true);
			fclose(m_Output);
		}
	}
protected:

	virtual void DoWriteProlog() = 0;
	virtual void DoWriteRead(MappedRead const * const read) = 0;
	virtual void DoWritePair(MappedRead const * const read1, MappedRead const * const read2) = 0;
	virtual void DoWriteUnmappedRead(MappedRead const * const read) = 0;
	virtual void DoWriteEpilog() = 0;

	int Print(const char *format, ...) {
		va_list arg;
		int done;

		va_start(arg, format);
		done = vsprintf(writeBuffer + bufferPosition, format, arg);
		bufferPosition += done;
		va_end(arg);
		return done;
	}

	void Flush(bool last = false) {
		if (bufferPosition > BUFFER_LIMIT || last) {
			fwrite(writeBuffer, sizeof(char), bufferPosition, m_Output);
			bufferPosition = 0;
			fflush(m_Output);
		}
	}

	bool paired_end;
	int min_dist;
	int max_dist;
	float identity;

//	bool IsProperPaired(MappedRead const * const read) {
//		read->TLS();
//		bool same_chr = (read->TLS()->Location.m_RefId == read->Paired->TLS()->Location.m_RefId);
//
//		//TODO: Fix. m_Location is now unsigned!
////		throw "Fix this!";
//		int dist = abs((long) read->TLS()->Location.m_Location - (long) read->Paired->TLS()->Location.m_Location);
//
//		return ((read->Identity > identity && read->Paired->Identity > identity) && (read->Paired->Strand != read->Strand) && same_chr
//				&& (dist < max_dist && dist > min_dist));
//	}

private:
	FILE * m_Output;

	static int const BUFFER_SIZE = 17000000;
	static int const BUFFER_LIMIT = 16000000;

	char * writeBuffer;
	int bufferPosition;

public:
	void WriteProlog() {
		if (m_Output)
			DoWriteProlog();
	}

	void WriteRead(MappedRead const * const read, bool mapped = true) {
		if (m_Output) {
			if (mapped) {
				DoWriteRead(read);
			} else {
				DoWriteUnmappedRead(read);
			}
			Flush();
		}
	}

	void WritePair(MappedRead const * const read1, MappedRead const * const read2) {
		if (m_Output) {
			DoWritePair(read1, read2);
			Flush();
		}
	}

//	void WriteRead(std::list<MappedRead const * const> reads) {
//		if (m_Output) {
//			for(int i = 0; i < reads.size(); ++i) {
//				WriteRead(reads[i]);
//			}
//		}
//	}
	void WriteEpilog() {
		if (m_Output)
			DoWriteEpilog();
	}
};

#endif
