#include "SW.h"

#include "Debug.h"
#include "Timing.h"

ulong SW::scoreCount = 0;

static int refMaxLen = 0;
static int qryMaxLen = 0;

long tCount = 1;
long tSum = 0;
long brokenPairs = 0;

void SW::DoRun() {
	int const corridor = Config.GetInt("corridor");

	int batchSize = NGM.Aligner()->GetScoreBatchSize();

	m_QryBuffer = new char const *[batchSize];
	m_RefBuffer = new char const *[batchSize];
	m_ScoreBuffer = new float[batchSize];

	char * m_DirBuffer = new char[batchSize];

	bool m_EnableBS = false;
	if (Config.Exists("bs_mapping"))
		m_EnableBS = (Config.GetInt("bs_mapping", 0, 1) == 1);

	qryMaxLen = Config.GetInt("qry_max_len");
	refMaxLen = ((qryMaxLen + Config.GetInt("corridor")) | 1) + 1;

	for (int i = 0; i < batchSize; ++i) {
		m_RefBuffer[i] = new char[refMaxLen];
	}

	LocationScore * * scores = new LocationScore*[batchSize];

	//Timer x;
	//x.ST();
	while (NGM.ThreadActive(m_TID, GetStage()) && (NGM.StageActive(GetStage() - 1) || NGM.bCSSW.Count() > 0)) {
		bool resume = true;

		if (resume) {
			int count = NGM.bCSSW.Read(scores, batchSize);

			if (count > batchSize) {
				Log.Error("returned count=%i > batchSize=%i", count, batchSize);
				Fatal();
			}

			if (count != 0) {
				for (int i = 0; i < count; ++i) {
					MappedRead * cur_read = scores[i]->Read;
					SequenceLocation loc = scores[i]->Location;
					if (NGM.DualStrand() && (loc.m_RefId & 1)) {
						// RefId auf +-Strang setzen
						--loc.m_RefId;
						cur_read->computeReverseSeq();
						m_QryBuffer[i] = cur_read->RevSeq;
						m_DirBuffer[i] = 1;
					} else {
						m_QryBuffer[i] = cur_read->Seq;
						m_DirBuffer[i] = 0;
					}

					SequenceProvider.DecodeRefSequence(const_cast<char *>(m_RefBuffer[i]), loc.m_RefId, loc.m_Location - (corridor >> 1),
							refMaxLen);

					m_ScoreBuffer[i] = -1;

//					Log.Message("%u %u %u %u", loc.m_Location, corridor, (corridor >> 1), loc.m_Location - (corridor >> 1));
//					Log.Message("Strand: %c", cur_read->Strand);
//					Log.Message("Ref:  %.*s", refMaxLen, m_RefBuffer[i]);
//					Log.Message("Read: %s", m_QryBuffer[i]);
				}
//				Log.Message("SW Thread %i launching batch (size=%i)", m_TID, count);
				Timer tmr;
				tmr.ST();
				SW::scoreCount += count;
				int res = NGM.Aligner()->BatchScore(m_AlignMode, count, m_RefBuffer, m_QryBuffer, m_ScoreBuffer, (m_EnableBS) ? m_DirBuffer : 0);
				if (res != count)
					Log.Error("SW Kernel couldnt calculate all scores (%i out of %i)", res, count);

//				Log.Warning("SW Thread %i finished batch (Size = %i, Elapsed: %.2fs)", m_TID, count, tmr.ET());
				Timer y;
				y.ST();
				brokenPairs = 0;
				for (int i = 0; i < count; ++i) {
					scores[i]->Score.f = m_ScoreBuffer[i];
					if (AtomicInc(&(scores[i]->Read->Calculated)) == scores[i]->Read->nScores()) {
						SendToPostprocessing(scores[i]->Read);
					}
				}
				NGM.Stats->brokenPairs = brokenPairs * 100.0f / count;
				NGM.Stats->insertSize = tSum / tCount;
//				Log.Warning("Finished post processing in %.2fs", y.ET());
			} else {
				Log.Verbose("Nothing to do...waiting");
				Sleep(100);
			} // if (count != 0)
		} // if (resume)
		  //x.ST();
	} // while

//	CommitReads();

	delete[] m_DirBuffer;
	m_DirBuffer = 0;
	delete[] scores;
	scores = 0;
	delete[] m_ScoreBuffer;
	m_ScoreBuffer = 0;

	for (int i = 0; i < batchSize; ++i) {
		delete[] m_RefBuffer[i];
		m_RefBuffer[i] = 0;
	}

	delete[] m_RefBuffer;
	m_RefBuffer = 0;
	delete[] m_QryBuffer;
	m_QryBuffer = 0;

	NGM.bSWO.Release();
	FinishStage();

	Log.Verbose("SW Thread %i finished", m_TID);
}

//void SW::SendToPostprocessing(MappedRead * read) {
//	delete read; read = 0;
//}

long count = 0;
long sum = 0;

void SW::SendToPostprocessing(MappedRead * read) {
	//if (!NGM.LastPartition()) {
	//	NGM.BufferRead(read);
	//} else {
//	if (read->Scores.size() == 0) {
//		NGM.AddUnmappedRead(read->ReadId, MFAIL_NOCAND);
//		Log.Verbose("Read %i not mapped (no candidates/scores)", read->ReadId);
//		//TODO: add call to GenericWriter::writeUnmappedRead()
//		delete read;
//		return;
//	}

	// Program runs in Paired mode and current read got a pair
	if (NGM.Paired() && (!read->HasFlag(NGMNames::PairedFail))) {
		if (read->Paired == 0) {
			Log.Error("No read pair found.");
			Fatal();
		}
		Log.Verbose("[PAIRED] SW::SendToPostprocessing: %i (%s)", read->ReadId, read->name);
		// Paired read finished score calculation
		// -> race condition here: both paired reads could enter at the same time
		//Log.Message("%d %d", read->Paired->Calculated, read->Paired->nScores());
		if (read->Paired->Calculated > -1 && read->Paired->Calculated == read->Paired->nScores()) {
			if (read->ReadId < read->Paired->ReadId) {
				if (AtomicInc(&read->Lock) == 1) {
					PairedReadSelection(read, read->Paired);
				}
			} else {
				if (AtomicInc(&read->Paired->Lock) == 1) {
					PairedReadSelection(read->Paired, read);
				}
			}
		}
	} else {
		Log.Verbose("[SINGLE] SW::SendToPostprocessing: %i (%s)", read->ReadId, read->name);
		if (!read->hasCandidates()) {
			//Read has no candidates. Nevertheless, write to buffer -> Output.cpp will check
			//Scores.size() and write it as unmapped to output.
			//NGM.bSWO.Write(&read, 1);
			//NGM.AddUnmappedRead(read, MFAIL_NOCAND);
			//Log.Verbose("Read %s (%i) not mapped (no candidates/scores)", read->name, read->ReadId);
			//NGM.GetReadProvider()->DisposeRead(read);

			//Log.Verbose("Deleting read %d (THIS SHOULD NOT HAPPEN).", cur_read->ReadId);
			NGM.AddUnmappedRead(read, MFAIL_NOCAND);
			NGM.SaveRead(read, false);

			Log.Verbose("Read %s (%i) not mapped (no candidates/scores)", read->name, read->ReadId);
			NGM.GetReadProvider()->DisposeRead(read);
			return;
		}

		float score_max = 0;
		float score_smax = 0;
		int score_max_loc = 0;
		int score_max_count = 0;

		//TODO: compute mapping quality based on score list
		//read->mappingQlty = 60;

		// zähle anzahl topscores
		for (int j = 0; j < read->nScores(); ++j) {
//			Log.Verbose("Candidate %i: Loc %i, Score %f",
			//j, read->Scores[j]->Location.m_Location, read->Scores[j]->Score.f
			//);
			if (read->Scores[j]->Score.f > score_smax) {
				if (read->Scores[j]->Score.f > score_max) {
					score_smax = score_max;
					score_max = read->Scores[j]->Score.f;
					score_max_loc = j;
					score_max_count = 1;
				} else if (read->Scores[j]->Score.f == score_max){
					++score_max_count;
					score_smax = score_max;
				} else {
					score_smax = read->Scores[j]->Score.f;
				}
			} else if (read->Scores[j]->Score.f == score_max) {
				++score_max_count;
			}

//			if (read->Scores[j]->Score.f > score_max) {
//				score_max = read->Scores[j]->Score.f;
//				score_max_loc = j;
//				score_max_count = 1;
//			} else if (read->Scores[j]->Score.f == score_max) {
//				++score_max_count;
//			}
		}

//		float cutoff = score_max - (3 * 7);
//		int addAlign = 0;
//		float smax_score = 0.0f;
//		for (int j = 0; j < read->nScores(); ++j) {
//			if (read->Scores[j]->Score.f >= cutoff && score_max_loc != j) {
//				addAlign += 1;
//			}
//			if (read->Scores[j]->Score.f > smax_score && score_max_loc != j) {
//				smax_score = read->Scores[j]->Score.f;
//			}
//		}
		int mq = 60.0f * (score_max - score_smax) / score_max;
		//int mq = 60;
		//int mq2 = mq * (std::min(100, read->mappingQlty) / 100.0f);
//		Log.Message("%s: %f %f -> %d, (%d) => %d", read->name, score_max, score_smax, mq, read->mappingQlty, mq2);
		read->mappingQlty = mq;

		//Log.Message("%s: %d %d (%f %f)", read->name, addAlign, score_max_count, score_max, cutoff);

		static int const eslimit = Config.GetInt("max_equal");
		if (eslimit > 0 && score_max_count > eslimit) {
			Log.Verbose(
					"Read %i rejected for hitting equal scoring limit (%i top scores of %i)",
					read->ReadId, score_max_count, score_max);

			//If read has more than eslimit equal scoring positions, pick one candidate region randomly and set
			//mapping quality to 0
			score_max_count = 1;
			read->EqualScoringCount = 1;
			//read->mappingQlty = 0;
		}

		read->TopScore = score_max_loc;
		NGM.bSWO.Write(&read, 1);
		NGM.AddMappedRead(read->ReadId);

//		Log.Verbose("Read %i: Loc %i (Ref %i), Score %f",
		//read->ReadId, read->TLS()->Location.m_Location, read->TLS()->Location.m_RefId, read->TLS()->Score.f);

		if (score_max_count > 1) {
//			Log.Verbose("Read %i got %i Locations with max score %.2f like Location <%i %i>",
			//read->ReadId, score_max_count, score_max, read->Scores[score_max_loc]->Location.m_Location, read->Scores[score_max_loc]->Location.m_RefId);
			//read->mappingQlty = 0;
			read->EqualScoringCount = score_max_count;

			int esid = 1;
			for (int j = score_max_loc + 1; j < read->nScores(); ++j) {
				if (read->Scores[j]->Score.f == score_max) {
//					Log.Verbose("Read %i got max Score %.2f at Location <%i %i>",
//							read->ReadId, score_max, read->Scores[j]->Location.m_Location, read->Scores[j]->Location.m_RefId);
					MappedRead * ntopread = new MappedRead(read->ReadId);
					ntopread->EqualScoringID = esid++;
					ntopread->EqualScoringCount = score_max_count;
					ntopread->Calculated = 1;
					ntopread->TopScore = 0;

					ntopread->Seq = new char[qryMaxLen];

//TODO: only copy pointer, faster
					memcpy(ntopread->Seq, read->Seq, qryMaxLen);
					if (read->RevSeq != 0) {
						ntopread->RevSeq = new char[qryMaxLen];
						memcpy(ntopread->RevSeq, read->RevSeq, qryMaxLen);
					}
					ntopread->qlty = 0;
					if (read->qlty != 0) {
						ntopread->qlty = new char[qryMaxLen];
						memcpy(ntopread->qlty, read->qlty, qryMaxLen);
					}

					ntopread->name = new char[read->nameLength];
					memcpy(ntopread->name, read->name, read->nameLength);
					ntopread->nameLength = read->nameLength;

					ntopread->length = read->length;
					ntopread->AddScore(*read->Scores[j])->Read = ntopread;
					NGM.bSWO.Write(&ntopread, 1);
//delete ntopread;
//m_ReadBuffer.push_back(ntopread);
				}
			}
		}
		//delete read;
	}
	//}
}

bool SW::CheckPairs(LocationScore * ls1, LocationScore * ls2, float & topScore, int & dst, bool & equalScore) {
	if (ls1->Location.m_RefId == (ls2->Location.m_RefId ^ 0x1)) {
		int distance =
				(ls2->Location.m_Location > ls1->Location.m_Location) ?
						ls2->Location.m_Location - ls1->Location.m_Location : ls1->Location.m_Location - ls2->Location.m_Location;

		//Log.Message("%d %d", _NGM::sPairMinDistance, _NGM::sPairMaxDistance);
//		if(distance < 0) {
//			Log.Error("here!!");
//		}

		if (distance > _NGM::sPairMinDistance && distance < _NGM::sPairMaxDistance) {
//			Log.Green("[%d, %d] Score: %f, Distance: %d (%u - %u), AVG: %d", ls1->Read->ReadId, ls2->Read->ReadId, ls1->Score.f + ls2->Score.f, distance, ls2->Location.m_Location, ls1->Location.m_Location, tSum / tCount);
			float f = ls1->Score.f + ls2->Score.f;
			if (f > topScore) {
				topScore = f;
				dst = distance;
				return true;
			} else if (f == topScore) {
				equalScore = true;
				Log.Verbose("Found a pair (%d, %d) with same score (%f). Choosing pair according to insert size.", ls1->Read->ReadId, ls2->Read->ReadId, f);
				int avg = tSum / tCount;
				if (abs(dst - avg) > abs(distance - avg)) {
					topScore = f;
					dst = distance;
					return true;
				} else if (abs(dst) == abs(distance)) {
					Log.Verbose("Same insert size (%d). Choosing randomly.", distance);
				}
			}
		}
	}
	return false;
}

void SW::PairedReadSelection(MappedRead * read1, MappedRead * read2) {
	Log.Verbose("Paired Read selection -> read %i: %i candidates / read %i: %i candidates", read1->ReadId, read1->nScores(), read2->ReadId, read2->nScores());

	if (read1->HasFlag(NGMNames::PairedFail) || read2->HasFlag(NGMNames::PairedFail))
		Log.Error("Paired read selection recursion (read %i, %i)", read1->ReadId, read2->ReadId);

	for (int i = 0; i < read1->nScores(); ++i) {
		Log.Verbose("Read %i c%i -> Location <%i, %i>, Score %f", read1->ReadId, i,
				read1->Scores[i]->Location.m_Location, read1->Scores[i]->Location.m_RefId,
				read1->Scores[i]->Score.f);
	}

	for (int i = 0; i < read2->nScores(); ++i) {
		Log.Verbose("Read %i c%i -> Location <%i, %i>, Score %f", read2->ReadId, i,
				read2->Scores[i]->Location.m_Location, read2->Scores[i]->Location.m_RefId,
				read2->Scores[i]->Score.f); // */
	}

	float topScore = 0;
	int distance = 0;
	bool equalScoreFound = false;
	for (int i = 0; i < read1->nScores(); ++i) {
		for (int j = 0; j < read2->nScores(); ++j) {
			if (CheckPairs(read1->Scores[i], read2->Scores[j], topScore, distance, equalScoreFound)) {
				read1->TopScore = i;
				read2->TopScore = j;
			}
		}
	}

	if (read1->TopScore != -1) {
		Log.Verbose("Read pairing found: R %i (#%i) at <%i, %i> Score %f, R %i (#%i) at <%i, %i> Score %f / Distance %i",
				read1->ReadId, read1->TopScore, read1->Scores[read1->TopScore]->Location.m_Location,
				read1->Scores[read1->TopScore]->Location.m_RefId, read1->Scores[read1->TopScore]->Score.f,
				read2->ReadId, read2->TopScore, read2->Scores[read2->TopScore]->Location.m_Location,
				read2->Scores[read2->TopScore]->Location.m_RefId, read2->Scores[read2->TopScore]->Score.f,
				distance);

		//TODO: compute quality
		read1->mappingQlty = read2->mappingQlty = 60;
		if (equalScoreFound) {
			read1->mappingQlty = read2->mappingQlty = 0;
		}

		NGM.bSWO.Write(&read1, 1);
		NGM.bSWO.Write(&read2, 1);
		//delete read1;
		//delete read1;
		NGM.AddMappedRead(read1->ReadId);
		NGM.AddMappedRead(read2->ReadId);

		tCount += 1;
		tSum += distance;

//		Log.Message("%s/%s: %d (%d)", read1->name, read2->name, distance, tSum / tCount);

	} else {
		read1->SetFlag(NGMNames::PairedFail);
		//NGM.AddUnmappedRead(read2, MFAIL_PAIRING);
		read2->SetFlag(NGMNames::PairedFail);

		//delete read1;
		//delete read2;

		Log.Verbose("Broken pair: %i, %i (%s)", read1->ReadId, read2->ReadId, read1->name);
		brokenPairs += 1;

		SendToPostprocessing(read1);
		SendToPostprocessing(read2);

	}
}

//bool SW::CommitReads() {
//	int written = 0;
//	//Log.Message("SW Thread %i committing finished reads", m_TID);
////	while (!m_ReadBuffer.empty()) {
////		MappedRead * read = m_ReadBuffer.front();
////		NGM.bSWO.Write(&read, 1);
////		//delete read;
////		m_ReadBuffer.pop_front();
////		++written;
////	}
////	Log.Message("SW Thread %i finished commit (%i Reads written)", m_TID, written);
//	return true;
//}
